@@ -6,15 +6,20 @@ from pgtrigger.core import disable
 from pgtrigger.core import enable
 from pgtrigger.core import F
 from pgtrigger.core import get
+from pgtrigger.core import ignore
 from pgtrigger.core import Insert
 from pgtrigger.core import install
+from pgtrigger.core import InsteadOf
 from pgtrigger.core import IsDistinctFrom
 from pgtrigger.core import IsNotDistinctFrom
 from pgtrigger.core import Protect
 from pgtrigger.core import prune
 from pgtrigger.core import Q
+from pgtrigger.core import Referencing
 from pgtrigger.core import register
+from pgtrigger.core import Row
 from pgtrigger.core import SoftDelete
+from pgtrigger.core import Statement
 from pgtrigger.core import Trigger
 from pgtrigger.core import Truncate
 from pgtrigger.core import uninstall
@@ -33,15 +38,20 @@ __all__ = [
     'enable',
     'F',
     'get',
+    'ignore',
     'Insert',
     'install',
+    'InsteadOf',
     'IsDistinctFrom',
     'IsNotDistinctFrom',
     'Protect',
     'prune',
     'Q',
+    'Referencing',
     'register',
+    'Row',
     'SoftDelete',
+    'Statement',
     'Trigger',
     'Truncate',
     'uninstall',
