@@ -2,6 +2,7 @@ import contextlib
 import copy
 import hashlib
 import logging
+import threading
 
 from django.db import connection
 from django.db import models
@@ -9,13 +10,54 @@ from django.db.models.expressions import Col
 from django.db.models.fields.related import RelatedField
 from django.db.models.sql import Query
 from django.db.models.sql.datastructures import BaseTable
+import pgconnection
 
 
 LOGGER = logging.getLogger('pgtrigger')
 
 
 # All registered triggers for each model
-registry = set()
+registry = {}
+
+# All triggers currently being ignored
+_ignore = threading.local()
+
+
+def _is_concurrent_statement(sql):
+    """
+    True if the sql statement is concurrent and cannot be ran in a transaction
+    """
+    sql = sql.strip().lower() if sql else ''
+    return sql.startswith('create') and 'concurrently' in sql
+
+
+def _inject_pgtrigger_ignore(sql, sql_vars, cursor):  # pragma: no cover
+    """
+    A pgconnection pre_execute hook that sets a pgtrigger.ignore
+    variable in the executed SQL. This lets other triggers know when
+    they should ignore execution
+    """
+    if cursor.name:
+        # A named cursor automatically prepends
+        # "NO SCROLL CURSOR WITHOUT HOLD FOR" to the query, which
+        # causes invalid SQL to be generated. There is no way
+        # to override this behavior in psycopg2, so context tracking
+        # is ignored for named cursors. Django only names cursors
+        # for iterators and other statements that read the database,
+        # so it seems to be safe to ignore named cursors.
+        # TODO(@wesleykendall): Find a way to generate valid SQL
+        # for local variables within a named cursor declaration.
+        return None
+    elif _is_concurrent_statement(sql):
+        # Concurrent index creation is incompatible with local variable
+        # setting. Ignore this specific statement for now
+        return None
+
+    sql = (
+        'SET LOCAL pgtrigger.ignore=\'{' + ','.join(_ignore.value) + '}\';'
+    ) + sql
+
+    return sql, sql_vars
 
 
 def register(*triggers):
@@ -32,6 +74,45 @@ def register(*triggers):
     return _model_wrapper
 
 
+class _Level:
+    def __init__(self, name):
+        self.name = name
+
+    def __str__(self):
+        return self.name
+
+
+#: For specifying row-level triggers (the default)
+Row = _Level('ROW')
+
+#: For specifying statement-level triggers
+Statement = _Level('STATEMENT')
+
+
+class Referencing:
+    """For specifying the REFERENCING construct of a statement-level trigger"""
+
+    def __init__(self, *, old=None, new=None):
+        if not old and not new:
+            raise ValueError(
+                'Must provide either "old" and/or "new" to the referencing'
+                ' construct of a trigger'
+            )
+
+        self.old = old
+        self.new = new
+
+    def __str__(self):
+        ref = 'REFERENCING'
+        if self.old:
+            ref += f' OLD TABLE AS {self.old} '
+
+        if self.new:
+            ref += f' NEW TABLE AS {self.new} '
+
+        return ref
+
+
 class _When:
     def __init__(self, name):
         self.name = name
@@ -46,6 +127,9 @@ Before = _When('BEFORE')
 #: For specifying "AFTER" in the "when" clause of a trigger
 After = _When('AFTER')
 
+#: For specifying "INSTEAD OF" in the "when" clause of a trigger
+InsteadOf = _When('INSTEAD OF')
+
 
 class _Operation:
     def __init__(self, name):
@@ -234,9 +318,9 @@ class Q(models.Q, Condition):
         return sql
 
 
-def _drop_trigger(table, trigger_name):
+def _drop_trigger(table, trigger_pgid):
     with connection.cursor() as cursor:
-        cursor.execute(f'DROP TRIGGER IF EXISTS {trigger_name} ON {table};')
+        cursor.execute(f'DROP TRIGGER IF EXISTS {trigger_pgid} ON {table};')
 
 
 # Allows Trigger methods to be used as context managers, mostly for
@@ -253,19 +337,36 @@ class Trigger:
     creating derived trigger classes.
     """
 
+    name = None
+    level = Row
     when = None
     operation = None
     condition = None
+    referencing = None
     func = None
 
     def __init__(
-        self, *, when=None, operation=None, condition=None, func=None
+        self,
+        *,
+        name=None,
+        level=None,
+        when=None,
+        operation=None,
+        condition=None,
+        referencing=None,
+        func=None,
     ):
+        self._name = name
+        self.level = level or self.level
         self.when = when or self.when
         self.operation = operation or self.operation
         self.condition = condition or self.condition
+        self.referencing = referencing or self.referencing
         self.func = func or self.func
 
+        if not self.level or not isinstance(self.level, _Level):
+            raise ValueError(f'Invalid "level" attribute: {self.level}')
+
         if not self.when or not isinstance(self.when, _When):
             raise ValueError(f'Invalid "when" attribute: {self.when}')
 
@@ -274,6 +375,11 @@ class Trigger:
                 f'Invalid "operation" attribute: {self.operation}'
             )
 
+        if self.level == Row and self.referencing:
+            raise ValueError(
+                'Row-level triggers cannot have a "referencing" attribute'
+            )
+
     def __str__(self):
         return self.name
 
@@ -281,15 +387,30 @@ class Trigger:
         return hash(self.name)
 
     def get_key(self):
-        """The unique key for the trigger when installing"""
+        """The unique key for the trigger when generating an alias"""
         return list(self.__dict__.values())
 
     @property
     def name(self):
-        hash = hashlib.sha1(
-            ''.join(str(k) for k in self.get_key()).encode()
-        ).hexdigest()[:16]
-        return f'pgtrigger_{self.__class__.__name__.lower()}_'[:55] + str(hash)
+        if not self._name:
+            hash = hashlib.sha1(
+                ''.join(str(k) for k in self.get_key()).encode()
+            ).hexdigest()[:16]
+            return f'{self.__class__.__name__.lower()}_'[:37] + str(hash)
+        else:
+            return self._name
+
+    @property
+    def pgid(self):
+        """The ID of the trigger and function object in postgres
+
+        All objects are prefixed with "pgtrigger_" in order to be
+        discovered/managed by django-pgtrigger
+        """
+        if len(self.name) > 53:
+            raise ValueError(f'Trigger name "{self.name}" > 53 characters. ')
+
+        return f'pgtrigger_{self.name}'
 
     def get_condition(self, model):
         return self.condition
@@ -312,23 +433,30 @@ class Trigger:
         """
         if not self.func:
             raise ValueError(
-                'Must define func attribute or implement' ' get_func'
+                'Must define func attribute or implement get_func'
             )
         return self.func
 
-    def register(self, *model_classes):
+    def get_uri(self, model):
+        """The URI for the trigger in the registry"""
+        return (
+            f'{model._meta.app_label}.{model._meta.object_name}'
+            f':{self.name}'
+        )
+
+    def register(self, *models):
         """Register model classes with the trigger"""
-        for model_class in model_classes:
-            registry.add((model_class, self))
+        for model in models:
+            registry[self.get_uri(model)] = (model, self)
 
-        return _cleanup_on_exit(lambda: self.unregister(*model_classes))
+        return _cleanup_on_exit(lambda: self.unregister(*models))
 
-    def unregister(self, *model_classes):
+    def unregister(self, *models):
         """Unregister model classes with the trigger"""
-        for model_class in model_classes:
-            registry.remove((model_class, self))
+        for model in models:
+            del registry[self.get_uri(model)]
 
-        return _cleanup_on_exit(lambda: self.register(*model_classes))
+        return _cleanup_on_exit(lambda: self.register(*models))
 
     def render_condition(self, model):
         """Renders the condition SQL in the trigger declaration"""
@@ -354,13 +482,28 @@ class Trigger:
 
         return rendered_declare
 
+    def render_ignore(self, model):
+        """
+        Renders the clause that can dynamically ignore the trigger's execution
+        """
+        return '''
+            IF (_pgtrigger_should_ignore(TG_TABLE_NAME, TG_NAME) IS TRUE) THEN
+                IF (TG_OP = 'DELETE') THEN
+                    RETURN OLD;
+                ELSE
+                    RETURN NEW;
+                END IF;
+            END IF;
+        '''
+
     def render_func(self, model):
         """Renders the trigger function SQL statement"""
         return f'''
-            CREATE OR REPLACE FUNCTION {self.name}()
+            CREATE OR REPLACE FUNCTION {self.pgid}()
             RETURNS TRIGGER AS $$
                 {self.render_declare(model)}
                 BEGIN
+                    {self.render_ignore(model)}
                     {self.get_func(model)}
                 END;
             $$ LANGUAGE plpgsql;
@@ -371,9 +514,11 @@ class Trigger:
         table = model._meta.db_table
         return f'''
             DO $$ BEGIN
-                CREATE TRIGGER {self.name}
+                CREATE TRIGGER {self.pgid}
                     {self.when} {self.operation} ON {table}
-                    FOR EACH ROW {self.render_condition(model)} EXECUTE PROCEDURE {self.name}();
+                    {self.referencing or ''}
+                    FOR EACH {self.level} {self.render_condition(model)}
+                    EXECUTE PROCEDURE {self.pgid}();
             EXCEPTION
                 -- Ignore issues if the trigger already exists
                 WHEN others THEN null;
@@ -383,6 +528,9 @@ class Trigger:
     def install(self, model):
         """Installs the trigger for a model"""
 
+        # Ensure we have the function to ignore execution of triggers
+        install_ignore_func()
+
         rendered_func = self.render_func(model)
         rendered_trigger = self.render_trigger(model)
 
@@ -394,7 +542,7 @@ class Trigger:
 
     def uninstall(self, model):
         """Uninstalls the trigger for a model"""
-        _drop_trigger(model._meta.db_table, self.name)
+        _drop_trigger(model._meta.db_table, self.pgid)
 
         return _cleanup_on_exit(  # pragma: no branch
             lambda: self.install(model)
@@ -404,7 +552,7 @@ class Trigger:
         """Enables the trigger for a model"""
         with connection.cursor() as cursor:
             cursor.execute(
-                f'ALTER TABLE {model._meta.db_table} ENABLE TRIGGER {self.name};'
+                f'ALTER TABLE {model._meta.db_table} ENABLE TRIGGER {self.pgid};'
             )
 
         return _cleanup_on_exit(  # pragma: no branch
@@ -415,13 +563,41 @@ class Trigger:
         """Disables the trigger for a model"""
         with connection.cursor() as cursor:
             cursor.execute(
-                f'ALTER TABLE {model._meta.db_table} DISABLE TRIGGER {self.name};'
+                f'ALTER TABLE {model._meta.db_table} DISABLE TRIGGER {self.pgid};'
             )
 
         return _cleanup_on_exit(  # pragma: no branch
             lambda: self.enable(model)
         )
 
+    @contextlib.contextmanager
+    def ignore(self, model):
+        """Ignores the trigger in a single thread of execution"""
+        with contextlib.ExitStack() as pre_execute_hook:
+
+            # Create the table name / trigger name URI to pass down to the
+            # trigger.
+            ignore_uri = f'{model._meta.db_table}:{self.pgid}'
+
+            if not hasattr(_ignore, 'value'):
+                _ignore.value = set()
+
+            if not _ignore.value:
+                # If this is the first time we are ignoring trigger execution,
+                # register the pre_execute_hook
+                pre_execute_hook.enter_context(
+                    pgconnection.pre_execute_hook(_inject_pgtrigger_ignore)
+                )
+
+            if ignore_uri not in _ignore.value:
+                try:
+                    _ignore.value.add(ignore_uri)
+                    yield
+                finally:
+                    _ignore.value.remove(ignore_uri)
+            else:  # The trigger is already being ignored
+                yield
+
 
 class Protect(Trigger):
     """A trigger that raises an exception"""
@@ -443,13 +619,13 @@ class SoftDelete(Trigger):
     operation = Delete
     field = None
 
-    def __init__(self, *, condition=None, field=None):
+    def __init__(self, *, name=None, condition=None, field=None):
         self.field = field or self.field
 
         if not self.field:  # pragma: no cover
             raise ValueError('Must provide "field" for soft delete')
 
-        super().__init__(condition=condition)
+        super().__init__(name=name, condition=condition)
 
     def get_func(self, model):
         soft_field = model._meta.get_field(self.field).column
@@ -462,27 +638,43 @@ class SoftDelete(Trigger):
         '''
 
 
-def get():
+def get(*uris):
     """
-    Get all triggers registered to models
+    Get triggers matching URIs or all triggers registered to models
 
-    Note: Triggers can also be added to models with the pgtrigger.config
-    decorator
+    A URI is in the format of "app_label.model_name:trigger_name"
     """
-    return registry
+    if uris:
+        for uri in uris:
+            if uri and len(uri.split(':')) == 1:
+                raise ValueError(
+                    'Trigger URI must be in the format of'
+                    ' "app_label.model_name:trigger_name"'
+                )
+            elif uri and uri not in registry:
+                raise ValueError(
+                    f'URI "{uri}" not found in pgtrigger registry'
+                )
+
+        return [registry[uri] for uri in uris]
+    else:
+        return list(registry.values())
 
 
-def install():
+def install(*uris):
     """
-    Install all triggers registered to models
+    Install registered triggers matching URIs or all triggers if URIs aren't
+    provided. If URIs aren't provided, prune any orphaned triggers from the
+    database
     """
-    for model, trigger in get():
+    for model, trigger in get(*uris):
         LOGGER.info(
             f'pgtrigger: Installing "{trigger}" trigger for {model._meta.db_table} table.'
         )
         trigger.install(model)
 
-    prune()
+    if not uris:  # pragma: no branch
+        prune()
 
 
 def prune():
@@ -492,7 +684,7 @@ def prune():
     it is removed from the database
     """
     installed = {
-        (model._meta.db_table, trigger.name) for model, trigger in get()
+        (model._meta.db_table, trigger.pgid) for model, trigger in get()
     }
 
     with connection.cursor() as cursor:
@@ -512,9 +704,10 @@ def prune():
             _drop_trigger(*trigger)
 
 
-def enable():
+def enable(*uris):
     """
-    Enables all registered triggers
+    Enables registered triggers matching URIs or all triggers if no URIs
+    are provided
     """
     for model, trigger in get():
         LOGGER.info(
@@ -523,9 +716,10 @@ def enable():
         trigger.enable(model)
 
 
-def uninstall():
+def uninstall(*uris):
     """
-    Uninstalls all registered triggers.
+    Uninstalls registered triggers matching URIs or all triggers if no
+    URIs are provided.
 
     Running migrations will re-install any existing triggers. This
     behavior is overridable with ``settings.PGTRIGGER_INSTALL_ON_MIGRATE``
@@ -533,7 +727,7 @@ def uninstall():
     Note: This will not uninstall triggers when deleting a model.
     This operation is performed by the "prune" command.
     """
-    for model, trigger in get():
+    for model, trigger in get(*uris):
         LOGGER.info(
             f'pgtrigger: Uninstalling "{trigger}" trigger for {model._meta.db_table} table.'
         )
@@ -542,12 +736,66 @@ def uninstall():
     prune()
 
 
-def disable():
+def disable(*uris):
     """
-    Disables all registered triggers
+    Disables registered triggers matching URIs or all triggers if no URIs are
+    provided
     """
-    for model, trigger in get():
+    for model, trigger in get(*uris):
         LOGGER.info(
             f'pgtrigger: Disabling "{trigger}" trigger for {model._meta.db_table} table.'
         )
         trigger.disable(model)
+
+
+@contextlib.contextmanager
+def ignore(*uris):
+    """
+    Dynamically ignore registered triggers matching URIs from executing in
+    an individual thread.
+    If no URIs are provided, ignore all pgtriggers from executing in an
+    individual thread.
+    """
+    with contextlib.ExitStack() as stack:
+        for model, trigger in get(*uris):
+            stack.enter_context(trigger.ignore(model))
+
+        yield
+
+
+def install_ignore_func():
+    """
+    pgtrigger uses a special postgres function to determine when a trigger
+    should be ignored. This installs the function.
+
+    This function is automatically installed when all triggers are installed
+    with pgtrigger.install()
+    """
+    with connection.cursor() as cursor:
+        cursor.execute(
+            '''
+            CREATE OR REPLACE FUNCTION _pgtrigger_should_ignore(
+                table_name NAME,
+                trigger_name NAME
+            )
+            RETURNS BOOLEAN AS $$
+                DECLARE
+                    _pgtrigger_ignore TEXT[];
+                    _result BOOLEAN;
+                BEGIN
+                    BEGIN
+                        SELECT INTO _pgtrigger_ignore
+                            CURRENT_SETTING('pgtrigger.ignore');
+                        EXCEPTION WHEN OTHERS THEN
+                    END;
+                    IF _pgtrigger_ignore IS NOT NULL THEN
+                        SELECT CONCAT(table_name, ':', trigger_name) = ANY(_pgtrigger_ignore)
+                        INTO _result;
+                        RETURN _result;
+                    ELSE
+                        RETURN FALSE;
+                    END IF;
+                END;
+            $$ LANGUAGE plpgsql;
+            '''
+        )
