@@ -9,6 +9,46 @@ import pgtrigger.core
 from pgtrigger.tests import models
 
 
+@pytest.mark.django_db
+def test_statement_row_level_logging():
+    """
+    Updates "ToLogModel" entries, which have statement, row-level,
+    and referencing statement triggers that create log entries.
+    """
+    ddf.G(models.ToLogModel, n=5, field='old_field')
+
+    assert not models.LogEntry.objects.exists()
+
+    models.ToLogModel.objects.update(field='new_field')
+
+    # The statement-level trigger without references should have produced
+    # one log entry
+    assert (
+        models.LogEntry.objects.filter(
+            level='STATEMENT', old_field__isnull=True
+        ).count()
+        == 1
+    )
+
+    # The statement-level trigger with references should have made log
+    # entries for all of the old values and the new updated values
+    assert (
+        models.LogEntry.objects.filter(
+            level='STATEMENT', old_field__isnull=False
+        ).count()
+        == 5
+    )
+    assert (
+        models.LogEntry.objects.filter(
+            level='STATEMENT', old_field='old_field', new_field='new_field'
+        ).count()
+        == 5
+    )
+
+    # The row-level trigger should have produced five entries
+    assert models.LogEntry.objects.filter(level='ROW').count() == 5
+
+
 @pytest.mark.django_db
 def test_soft_delete():
     """
@@ -198,40 +238,77 @@ def test_complex_conditions():
             test_model.save()
 
 
+def test_referencing_rendering():
+    """Verifies the rendering of the Referencing construct"""
+    assert (
+        str(pgtrigger.Referencing(old='old_table')).strip()
+        == 'REFERENCING OLD TABLE AS old_table'
+    )
+    assert (
+        str(pgtrigger.Referencing(new='new_table')).strip()
+        == 'REFERENCING NEW TABLE AS new_table'
+    )
+    assert (
+        str(pgtrigger.Referencing(old='old_table', new='new_table')).strip()
+        == 'REFERENCING OLD TABLE AS old_table  NEW TABLE AS new_table'
+    )
+
+
 def test_arg_checks():
     """
     There are quite a few places that check arguments in the trigger module.
     Enumerate these cases here to make sure they work
     """
 
+    with pytest.raises(
+        ValueError, match='Must provide either "old" and/or "new"'
+    ):
+        pgtrigger.Referencing()
+
     with pytest.raises(ValueError, match='Must provide SQL'):
         pgtrigger.Condition()
 
     with pytest.raises(ValueError, match='Must provide at least one'):
         pgtrigger.UpdateOf()
 
+    with pytest.raises(ValueError, match='Invalid "level"'):
+        pgtrigger.Trigger(level='invalid')
+
     with pytest.raises(ValueError, match='Invalid "when"'):
         pgtrigger.Trigger(when='invalid')
 
     with pytest.raises(ValueError, match='Invalid "operation"'):
         pgtrigger.Trigger(when=pgtrigger.Before, operation='invalid')
 
+    with pytest.raises(ValueError, match='Row-level triggers cannot have'):
+        pgtrigger.Trigger(
+            when=pgtrigger.Before,
+            operation=pgtrigger.Update,
+            referencing=pgtrigger.Referencing(old='old_table'),
+        )
+
     with pytest.raises(ValueError, match='Must define func'):
         pgtrigger.Trigger(
             when=pgtrigger.Before, operation=pgtrigger.Update
         ).get_func(None)
 
+    with pytest.raises(ValueError, match='> 53'):
+        pgtrigger.Trigger(
+            when=pgtrigger.Before, operation=pgtrigger.Update, name='1' * 54
+        ).pgid
+
 
 def test_registry():
     """
     Tests dynamically registering and unregistering triggers
     """
     # The trigger registry should already be populated with our test triggers
-    assert len(pgtrigger.core.registry) == 2
+    assert len(pgtrigger.core.registry) == 6
 
     # Add a trigger to the registry
     trigger = pgtrigger.Trigger(
         when=pgtrigger.Before,
+        name='my_aliased_trigger',
         operation=pgtrigger.Insert | pgtrigger.Update,
         func="RAISE EXCEPTION 'no no no!';",
     )
@@ -239,15 +316,23 @@ def test_registry():
     # Register/unregister in context managers. The state should be the same
     # at the end as the beginning
     with trigger.register(models.TestModel):
-        assert len(pgtrigger.core.registry) == 3
-        assert (models.TestModel, trigger) in pgtrigger.core.registry
+        assert len(pgtrigger.core.registry) == 7
+        assert f'tests.TestModel:{trigger.name}' in pgtrigger.core.registry
 
         with trigger.unregister(models.TestModel):
-            assert len(pgtrigger.core.registry) == 2
+            assert len(pgtrigger.core.registry) == 6
             assert (models.TestModel, trigger) not in pgtrigger.core.registry
 
-    assert len(pgtrigger.core.registry) == 2
+        # Try obtaining trigger by alias
+        assert pgtrigger.get('tests.TestModel:my_aliased_trigger')
+
+    assert len(pgtrigger.core.registry) == 6
     assert (models.TestModel, trigger) not in pgtrigger.core.registry
+    with pytest.raises(ValueError, match='not found'):
+        pgtrigger.get('tests.TestModel:my_aliased_trigger')
+
+    with pytest.raises(ValueError, match='must be in the format'):
+        pgtrigger.get('tests.TestMode')
 
 
 def test_operations():
@@ -303,6 +388,91 @@ def test_custom_trigger_definitions():
         with pytest.raises(InternalError, match='no no no!'):
             test_model.save(update_fields=['int_field'])
 
+    # Protect statement-level creates
+    trigger = pgtrigger.Trigger(
+        level=pgtrigger.Statement,
+        when=pgtrigger.Before,
+        operation=pgtrigger.Update,
+        func="RAISE EXCEPTION 'bad statement!';",
+    )
+    with trigger.install(models.TestTrigger):
+        with pytest.raises(InternalError, match='bad statement!'):
+            test_model.save()
+
+
+@pytest.mark.django_db(transaction=True)
+def test_basic_ignore():
+    """Verify basic dynamic ignore functionality"""
+    deletion_protected_model = ddf.G(models.TestTrigger)
+    with pytest.raises(InternalError, match='Cannot delete rows'):
+        deletion_protected_model.delete()
+
+    with pgtrigger.ignore('tests.TestTrigger:protect_delete'):
+        deletion_protected_model.delete()
+
+    assert not models.TestTrigger.objects.exists()
+
+    deletion_protected_model = ddf.G(models.TestTrigger)
+    with pytest.raises(InternalError, match='Cannot delete rows'):
+        deletion_protected_model.delete()
+
+
+@pytest.mark.django_db(transaction=True)
+def test_nested_ignore():
+    """Test nesting pgtrigger.ignore()"""
+    deletion_protected_model1 = ddf.G(models.TestTrigger)
+    deletion_protected_model2 = ddf.G(models.TestTrigger)
+    with pytest.raises(InternalError, match='Cannot delete rows'):
+        deletion_protected_model1.delete()
+
+    with pgtrigger.ignore('tests.TestTrigger:protect_delete'):
+        with pgtrigger.ignore('tests.TestTrigger:protect_delete'):
+            deletion_protected_model1.delete()
+        deletion_protected_model2.delete()
+
+    assert not models.TestTrigger.objects.exists()
+
+    deletion_protected_model = ddf.G(models.TestTrigger)
+    with pytest.raises(InternalError, match='Cannot delete rows'):
+        deletion_protected_model.delete()
+
+
+@pytest.mark.django_db(transaction=True)
+def test_multiple_ignores():
+    """Tests multiple pgtrigger.ignore()"""
+    deletion_protected_model1 = ddf.G(models.TestTrigger)
+    ddf.G(models.TestTrigger)
+    with pytest.raises(InternalError, match='Cannot delete rows'):
+        deletion_protected_model1.delete()
+
+    ddf.G(models.TestTrigger, field='hi!')
+    with pytest.raises(InternalError, match='no no no!'):
+        models.TestTrigger.objects.create(field='misc_insert')
+
+    with pgtrigger.ignore('tests.TestTrigger:protect_delete'):
+        deletion_protected_model1.delete()
+        with pytest.raises(InternalError, match='no no no!'):
+            models.TestTrigger.objects.create(field='misc_insert')
+
+        with pgtrigger.ignore('tests.TestTrigger:protect_misc_insert'):
+            m = models.TestTrigger.objects.create(field='misc_insert')
+            m.delete()
+
+        models.TestTrigger.objects.all().delete()
+
+    assert not models.TestTrigger.objects.exists()
+
+    deletion_protected_model = ddf.G(models.TestTrigger)
+    with pytest.raises(InternalError, match='Cannot delete rows'):
+        deletion_protected_model.delete()
+
+    # Ignore all triggers
+    with pgtrigger.ignore():
+        m = models.TestTrigger.objects.create(field='misc_insert')
+        models.TestTrigger.objects.all().delete()
+
+    assert not models.TestTrigger.objects.exists()
+
 
 @pytest.mark.django_db
 def test_protect():
